#!/usr/bin/env python3

from pathlib import Path
from xd_cwl_utils.classes.metadata.tool_metadata import ToolMetadata, ParentToolMetadata


def add_parent_tool(tool_name, tool_version, subtool_names=None, biotools_id=None):
    """
    Make the correct directory structure for adding a new command line tool. Optionally, create initialized CWL
    and metadata files. Run from cwl-tools directory.
    :param tool_name(str): Name of the tool
    :param tool_version(str): version of the tool
    :param subtool_names(list(str)): list of subtool names if the tool is broken into multiple subtools.
    :param mk_meta_files(bool): Specify whether to make initial CWL and metadata files.
    :return: None
    """
    tool_version = str(tool_version) # In case ArgumentParser is bypassed.

    common_dir = Path().cwd() / 'cwl-tools'/ tool_name / tool_version / 'common'
    common_dir.mkdir(parents=True)
    if biotools_id:
        parent_metadata = ParentToolMetadata.create_from_biotools(biotools_id, tool_version, subtool_names)
    else:
        parent_metadata = ParentToolMetadata(name=tool_name, softwareVersion=tool_version, featureList=subtool_names)
    new_file_path = common_dir / f"{parent_metadata.name}-metadata.yaml"
    parent_metadata.mk_file(new_file_path)
    return new_file_path


def add_tool(tool_name, tool_version, biotools_id=None):
    tool_version = str(tool_version)  # In case ArgumentParser is bypassed.
    new_tool_dir = Path.cwd() / 'cwl-tools' / tool_name / tool_version / tool_name
    new_tool_dir.mkdir(parents=True)
    if biotools_id:
        new_tool = ToolMetadata.create_from_biotools(biotools_id, tool_version)
    else:
        new_tool = ToolMetadata(name=tool_name, softwareVersion=tool_version)

    new_tool_path = new_tool_dir / f"{tool_name}-metadata.yaml"
    new_tool.mk_file(new_tool_path)
    return new_tool_path


def add_subtool(parent_rel_path, subtool_name):
    parent_path = Path(parent_rel_path) # Should be in /common
    parent_meta = ParentToolMetadata.load_from_file(parent_path)
    new_subtool_dir = parent_path.parents[1] / f"{parent_meta.name}_{subtool_name}"
    new_subtool_dir.mkdir()
    subtool_meta = parent_meta.make_subtool_metadata(subtool_name, parent_path)
    new_file_path = new_subtool_dir / f"{subtool_name}-metadata.yaml"
    subtool_meta.mk_file(new_file_path)
    return new_file_path



import tempfile
import logging
from pathlib import Path
from ruamel.yaml import safe_load
from semantic_version import Version
from .content_maps import make_tools_map, make_script_maps
from .validate_metadata import main as validate_meta
from .helpers.get_paths import get_metadata_path
from .helpers.validate_cwl import validate_cwl_tool, validate_cwl_tool_2
from .validate_inputs import validate_all_inputs_for_tool


def validate_tools_dir(cwl_tools_dir=None):
    """
    Validate all cwl files, metadata files, instances and instance metadata in a cwl-tools directory
    :return:
    """
    tool_map_temp_file = tempfile.NamedTemporaryFile(prefix='tools_map', suffix='.yaml', delete=True)  # Change to False if file doesn't persist long enough.
    make_tools_map(tool_map_temp_file.name, cwl_tools_dir=cwl_tools_dir)
    with tool_map_temp_file as tool_map:
        tool_map_dict = safe_load(tool_map)
    for identifier, values in tool_map_dict.items():
        tool_path = Path(values['path'])
        tool_type = values['type']
        file_version = Version(values['version'])


        if tool_type == 'parent':
            if not 'common' in tool_path.parts:
                raise ValueError  # an extra check.
            meta_type = 'parent_tool'  # correspond to command in validate_metadata
            validate_meta([meta_type, str(tool_path)])
            # assert no instances directory here?
        else:  # either a subtool or 'main' tool

            # validate metadata
            metadata_path = get_metadata_path(tool_path)
            meta_type = tool_type
            validate_meta([meta_type, str(metadata_path)])

            # validate cwl only if metadata specifies if it is version >= 1.0
            if file_version >= Version('1.0.0'):
                validate_cwl_tool_2(tool_path)

                # validate instances
                validate_all_inputs_for_tool(tool_path)

    return


def validate_scripts_dir(cwl_scripts_dir=None):

    script_map_temp_file = tempfile.NamedTemporaryFile(prefix='scripts_map', suffix='.yaml', delete=True)  # Change to False if file doesn't persist long enough.
    make_script_maps(script_map_temp_file.name, cwl_scripts_dir=cwl_scripts_dir)
    with script_map_temp_file as script_map:
        script_map_dict = safe_load(script_map)
    for identifier, values in script_map_dict.items():
        # validate metadata
        script_path = values['path']
        metadata_path = get_metadata_path(script_path)
        validate_meta(['script', str(metadata_path)])

        # validate cwl
        document_version = Version(values['version'])
        if document_version >= Version('1.0.0'):
            validate_cwl_tool_2(values['path'])
            validate_all_inputs_for_tool(script_path)
    return


def validate_repo(cwl_tools_dir=None, cwl_scripts_dir=None):
    validate_tools_dir(cwl_tools_dir=cwl_tools_dir)
    validate_scripts_dir(cwl_scripts_dir=cwl_scripts_dir)
    return



#
# * This file is subject to the terms and conditions defined in
# * file 'LICENSE.txt', which is part of this source code package.

from collections import OrderedDict
from pathlib import Path
import re
from abc import abstractmethod
from ruamel.yaml import safe_load
from ...classes.metadata.shared_properties import CodeRepository, WebSite, Person, Publication, Keyword, CallMap
from ...classes.metadata.common_functions import _mk_hashes
from ...classes.metadata.metadata_base import MetadataBase

class WorkflowMetadataBase(MetadataBase):
    @abstractmethod
    def _mk_identifier(self, **kwargs):
        pass

    @abstractmethod
    def _check_identifier(self, identifier):
        pass

    @property
    def identifier(self):
        return self._identifier

    @identifier.setter
    def identifier(self, identifier=None, **kwargs):
        if identifier:
            identifier = self._check_identifier(identifier)
        else:
            identifier = self._mk_identifier(**kwargs)
        self._identifier = identifier

    @property
    def keywords(self):
        return self._keywords

    @keywords.setter
    def keywords(self, keywords_list):
        if keywords_list:
            keywords = []
            for keyword in keywords_list:
                if isinstance(keyword, Keyword):
                    keywords.append(keyword)
                else:
                    if isinstance(keyword, dict):
                        keywords.append(Keyword(**keyword))
                    else:
                        keywords.append(Keyword(keyword))
        else:
            keywords = [Keyword()]
        self._keywords = keywords



class WorkflowMetadata(WorkflowMetadataBase):

    @staticmethod
    def _init_metadata():
        return OrderedDict([
        ('name', None),
        ('softwareVersion', None),
        ('description', None),
        ('identifier', None),
        ('version', '0.1'),
        ('callMap', [CallMap()]),
        ('codeRepository', CodeRepository()),
        ('WebSite', [WebSite()]),
        ('license', None),
        ('contactPoint', [Person()]),
        ('publication', [Publication()]),
        ('keywords', [Keyword()]),
        ('alternateName', None),
        ('creator', [Person()]),
        ('programmingLanguage', None),
        ('datePublished', None),
    ])

    def _check_identifier(self, identifier):
        if not identifier[:3] == "WF_":
            raise ValueError(f"Workflow identifiers must start with 'WF_' you provided {identifier}")
        else:
            hex_pattern = r'[0-9a-f]{6}\.[0-9a-f]{2}$'
            match_obj = re.match(hex_pattern, identifier[3:])
            if not match_obj:
                raise ValueError(f"Tool identifier not formatted correctly: {identifier}")
        return identifier

    def _mk_identifier(self, start=0):
        if not (self.name and self.softwareVersion):
            raise ValueError(f"Name and softwareVersion must be provided to make an identifier.")
        name_hash, version_hash = _mk_hashes(self.name, self.softwareVersion)
        identifier = f"WF_{name_hash[start:start + 6]}.{version_hash[:2]}"
        return identifier

    @classmethod
    def load_from_file(cls, file_path):
        file_path = Path(file_path)
        with file_path.open('r') as file:
            file_dict = safe_load(file)
        return cls(**file_dict)

    def make_instance(self):
        raise NotImplementedError


import os
import tempfile
from tests.test_base import TestBase
from xd_cwl_utils.config import config
from xd_cwl_utils.content_maps import make_tools_map, make_script_map, make_script_maps

class TestToolMaps(TestBase):

    def test_make_tools_map(self):
        outfile_path = config[os.environ.get('CONFIG_KEY')]['content_maps_dir'] / 'tool-maps.yaml'
        make_tools_map(outfile_path)
        return


    def test_make_script_map(self):
        group_name = 'ENCODE-DCC'
        project = 'atac-seq-pipeline'
        version = 'v1.1'
        make_script_map(group_name, project, version)

    def test_make_script_maps(self):

        make_script_maps(tempfile.NamedTemporaryFile().name)
        return


#!/usr/bin/env python3

# Module to call from command line.

import sys
import argparse
from src.add.add_tools import add_parent_tool, add_tool, add_subtool
from src.add.add_scripts import add_script, add_common_script_metadata
from src.add.add_workflows import add_workflow

parser = argparse.ArgumentParser(description='Initialize metadata and directories for a tool, script, or workflow.')
subparsers = parser.add_subparsers(description='Specify the command to run.', dest='command')

# add_tool parser
addtool = subparsers.add_parser('tool', help='add a new standalone tool.')
addtool.add_argument('tool_name', type=str, help="The name of the tool to add.")
addtool.add_argument('tool_version', type=str, help="The version of the tool to add.")
addtool.add_argument('--biotoolsID', type=str, help='biotools id from https://bio.tools')

# add_parent_tool parser
add_parent = subparsers.add_parser('parent_tool', help='add parent tool/common metadata')
add_parent.add_argument('tool_name', type=str, help='The name of the tool to add.')
add_parent.add_argument('tool_version', type=str, help="The version of the tool to add.")
add_parent.add_argument('subtools', help='list of subtools of the parent tool.', nargs='+')
add_parent.add_argument('--biotoolsID', type=str, help='biotools id from bio.tools')

# add_subtool parser
addsubtool = subparsers.add_parser('subtool', help='add a subtool. A parent must exist.')
addsubtool.add_argument('parent_path', help='The absolute or relative path of the metadata file that describes the parent of the subtool.')
addsubtool.add_argument('subtool_name', help="The name of the subtool. The subtool name must be present in the parent's featureList field.")

# add_common_script_parser
addscriptcommon = subparsers.add_parser('common_script', help='add script metadata that other scripts can inherit from')
addscriptcommon.add_argument('group_name', help='The name of the group directory that the script will go into.')
addscriptcommon.add_argument('project_name', help='The name of the project directory that the script will go into.')
addscriptcommon.add_argument('script_version', help='The version of the script.')
addscriptcommon.add_argument('filename', help="The name of common metadata file. File extensions and '-metadata.yaml' postfix must be omitted.")

# add_script parser
addscript = subparsers.add_parser('script', help='add a script')
addscript.add_argument('group_name', help='The name of the group directory that the script will go into.')
addscript.add_argument('project_name', help='The name of the project directory that the script will go into.')
addscript.add_argument('script_version', help='The version of the script.')
addscript.add_argument('script_name', help='The name of the script. File extensions should be omitted. Replace spaces with underscores')
addscript.add_argument('--parent_metadata', '-p', nargs='+', help="path(s) to common script metadata that the script should inherit from.")
# Could add flags for more kwargs here. Probably easiest to just add to file though.

addworkflow = subparsers.add_parser('workflow', help='add a workflow')
addworkflow.add_argument('group_name', help='The name of the group directory that the workflow will go into.')
addworkflow.add_argument('workflow_name', help='The name of the workflow. Replace spaces with underscores.')
addworkflow.add_argument('workflow_version', help='The version of the workflow.')


def main(args):
    if args.command == 'tool':
        add_tool(args.tool_name, args.tool_version, args.biotoolsID)
    elif args.command == 'parent_tool':
        add_parent_tool(args.tool_name, args.tool_version, args.subtools, args.biotoolsID)
    elif args.command == 'subtool':
        add_subtool(args.parent_path, args.subtool_name)
    elif args.command == 'common_script':
        add_common_script_metadata(args.group_name, args.project_name, args.script_version, args.filename)
    elif args.command == 'script':
        add_script(args.group_name, args.project_name, args.script_version, args.script_name, parentMetadata=args.parent_metadata)
    elif args.command == 'workflow':
        add_workflow(args.group_name, args.workflow_name, args.workflow_version)
    else:
        parser.print_help()
    return

if __name__ == "__main__":
    args = parser.parse_args()
    main(args)
    sys.exit(0)